#include <fcntl.h>
#include <assert.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <gnu/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <unistd.h>
#include <math.h>
#include <gtk/gtk.h>
#include <sys/soundcard.h>

#include "opledit.h"

static struct synth_info info;

#define SEQUENCER_PATH "/dev/sequencer"

static int seqfd;
static int sb_dev=-1;

static int num_voices, bender=0;
static int free_voices;

/* LRU list for free operators */
static unsigned char free_list[256];
static int fhead=0, ftail=0, flen=0;

/* LRU list for still playing notes */
static unsigned char note_list[256];

static int nhead=0, ntail=0, nlen=0;
static unsigned char oper_note[32];



SEQ_DEFINEBUF (2048);
SEQ_PM_DEFINES; /* patch manager stuff */


/*
 * The function seqbuf_dump() must always be provided
 */

void seqbuf_dump ()
{
  if (_seqbufptr)
    if (write (seqfd, _seqbuf, _seqbufptr) == -1)
      {
	perror ("write /dev/sequencer");
	exit (-1);
      }
  _seqbufptr = 0;
}


/* initialise note/operator lists, etc. */
static void init_lists()
{
  int i;
  for(i=0; i<16;i++) oper_note[i] = 255; /* free */
  /* this asumes that num_voices, set in init_sequencer(), is set. */
  flen = num_voices;
  for(i=0; i<num_voices; i++) {
    free_list[i]=i;
  };
  //bufp = 0;
};

/* initialise sequencer */

static void init_sequencer()
{
	int i, n;

	if((seqfd = open(SEQUENCER_PATH, O_WRONLY, 0)) == -1) {
		printf("Cannot open " SEQUENCER_PATH ".");
		exit(-1);
	};

	if (ioctl (seqfd, SNDCTL_SEQ_NRSYNTHS, &n) == -1) {
		printf("Cannot get " SEQUENCER_PATH " info.");
		exit(-1);
	};

	for (i = 0; i < n; i++) {
		info.device = i;
		if (ioctl (seqfd, SNDCTL_SYNTH_INFO, &info) == -1) {
			printf("Cannot get " SEQUENCER_PATH " info.");
			exit(-1);
		}

		if (info.synth_type == SYNTH_TYPE_FM
			&& info.synth_subtype == FM_TYPE_OPL3)
			sb_dev = i;
		//free_voices = num_voices = (info.nr_voices>999)?999:info.nr_voices;
		free_voices = num_voices = 6;
		//	printf("free_voices: %i\n",free_voices);
	}
	if(sb_dev == -1) {
		printf("OPL3 synth not found. (perhaps you have not loaded the kernel module.)");
		exit(-1);
	}
	if (ioctl (seqfd, SNDCTL_SEQ_RESET, &sb_dev) == -1)
		printf("Could not reset sequencer.");
	assert(ioctl(seqfd,SNDCTL_FM_4OP_ENABLE,&sb_dev)!=-1);
};


void load_fm(char *buf,bool ops) {

	struct sbi_instrument instr;

	int data_size;
	instr.channel = 1;
	instr.device = sb_dev;
	if(ops) {
		instr.key = OPL3_PATCH;
		data_size = 22;	
	}
	else  {
		instr.key = FM_PATCH;
		data_size = 11;
	}
	buf[10]|=0x30;
	buf[21]|=0x30;
	for (int n = 0; n < 32; n++)
		instr.operators[n] = (n < data_size) ? buf[n] : 0;
	SEQ_WRPATCH(&instr, sizeof(instr));
}



static void stop_note(int note, int velocity)
{
  int i, op=255;
  
  
  for (i=0;i<num_voices && op==255;i++) {
    if (oper_note[i]== note) op=i;
  }

  if (op==255) 
    {
      printf("Note %d off, note not started\n", note);
      printf("%d, %d\n", flen, nlen);
      return;	/* Has already been killed ??? */
    }
  
  SEQ_STOP_NOTE(sb_dev, op, note, velocity);
  SEQ_DUMPBUF();
  
  oper_note[op] = 255;
  
  free_list[ftail]=op;
  flen++;
  ftail = (ftail+1) % num_voices;
  
  for (i=0;i<16;i++)
    if (note_list[i] == op) note_list[i] = 255;
  
  while (nlen && note_list[nhead] == 255) {
      nlen--;
      nhead = (nhead+1) % 256;
    }
  free_voices++;
}

static void kill_one_note()
{
     int oldest;
     
     if (!nlen) {
	  printf("Free list empty but no notes playing\n"); 
	  return;
     }	/* No notes playing */
     
     oldest = note_list[nhead];
     nlen--;
     nhead = (nhead+1) % 256;
     
     printf("Killing oper %d, note %d\n", oldest, oper_note[oldest]);
     
     if (oldest== 255) 
	  return;	/* Was already stopped. Why? */
  
     stop_note(oper_note[oldest], 127);
}

static void start_note(int note, int velocity)
{
     int free;    
     /* for the csound output */
     //notetime[note] = sec() + rint(usec()/precision)*(precision/1e6);
     
     if (!flen) kill_one_note();
     
     if (!flen) {printf("** no free voices\n");return;}	/* Panic??? */
     
     free = free_list[fhead];
     flen--;
     fhead = (fhead+1) % num_voices;
     
     note_list[ntail] = free;
     
     if (nlen>255) {
	  nlen=0;	/* Overflow -> hard reset */
     }
     nlen++;
     ntail = (ntail+1) % 256;
     
     oper_note[free] = note;
     
     SEQ_SET_PATCH(sb_dev, free, 1);
     SEQ_PITCHBEND(sb_dev, free, bender);
     SEQ_START_NOTE(sb_dev, free, note, velocity);
     SEQ_DUMPBUF();
     free_voices--;
}

static void channel_pressure(int ch, int pressure)
{
     int i;
     for (i=0;i<num_voices;i++) {
	  if (oper_note[i] != 255) {
	       SEQ_CHN_PRESSURE(sb_dev, i, pressure);
	       SEQ_DUMPBUF();
	  }
     }
}

static void pitch_bender(int ch, int value)
{
     int i;
     value -= 8192;
     bender = value;
     for (i=0;i<num_voices;i++) {
	  if (oper_note[i] != 255) {
	       bender = value;
	       SEQ_PITCHBEND(sb_dev, i, value);
	       SEQ_DUMPBUF();
	  }
     }
}

void handle_channelop(unsigned char *buf)
{
     int ch = buf[0] & 0x0f;
     int value;
     
     switch (buf[0] & 0xf0) {
     case 0x90:	/* Note on */
	  if (buf[2])
	       start_note(buf[1], buf[2]);
	  else
	       stop_note(buf[1], buf[2]);
	  break;
	  
     case 0xb0:	/* Control change */
	  break;
	  
     case 0x80:	/* Note off */
	  stop_note(buf[1], buf[2]);
	  break;
	  
     case 0xe0:	/* Pitch bender */
	  value = ((buf[2] & 0x7f) << 7) | (buf[1] & 0x7f);
	  pitch_bender(ch, value);
	  break;
    
     case 0xc0:	// Pgm change 
	  break;
  
     case 0xd0:	/* Channel pressure */
	  channel_pressure(ch, buf[1]);
	  break;	  

     default:
	  break;
     }
}

static int midifd;

static void init_midi() {
	if(!strcmp(mididev,"-")) {
		midifd = STDIN_FILENO;
		fcntl(midifd,F_SETFL,O_NONBLOCK);
	}
	else {
		midifd = open(mididev,O_NONBLOCK|O_RDONLY);
		if(midifd<0) {
			perror(mididev);
			exit(1);
		}
	}
}
/*
static void handle_channelop(unsigned char *buf) {
	//fprintf(stderr,"channelop 0x%X\n",buf[0]);
	switch(buf[0]&0xf0) {
		case 0x80:
			//fprintf(stderr,"channelop 0x%X\n",buf[1]);
			if(buf[1]>=0x3c)
				buf[0] = buf[0]+1;
			buf[1]+=pitch_adjust;
			buf[2] = (unsigned char)(((int)buf[2]*127)/100);
			printf("Channel %i: ",(buf[0]&0x0f)+1);
			printf("Note Off: ");
			print_note(stdout,buf[1]);
			printf(" 0x%X\n",buf[2]);
			assert(write(fd,buf,3)==3);break;
		case 0x90:
			//fprintf(stderr,"channelop 0x%X\n",buf[1]);
			if(buf[1]>=0x3c)
				buf[0] = buf[0]+1;
			buf[1]+=pitch_adjust;
			buf[2] = (unsigned char)(((int)buf[2]*127)/100);
			printf("Channel %i: ",(buf[0]&0x0f)+1);
			printf("Note On: ");
			print_note(stdout,buf[1]);
			printf(" 0x%X\n",buf[2]);
			assert(write(fd,buf,3)==3);break;
		case 0xA0:
			buf[1]+=pitch_adjust;
			printf("Channel %i: ",(buf[0]&0x0f)+1);
			printf("Polymorphic Aftertouch: ");
			print_note(stdout,buf[1]);
			printf(" 0x%X\n",buf[2]);
			assert(write(fd,buf,3)==3);break;
		case 0xB0:
			printf("Channel %i: ",(buf[0]&0x0f)+1);
			printf("Control Change: 0x%X 0x%X\n",buf[1],buf[2]);
			assert(write(fd,buf,3)==3);break;
		case 0xC0:
			printf("Channel %i: ",(buf[0]&0x0f)+1);
			printf("Program Change: 0x%X\n",buf[1]);
			assert(write(fd,buf,2)==2);break;
		case 0xD0:
			printf("Channel %i: ",(buf[0]&0x0f)+1);
			printf("Channel Aftertouch: 0x%X\n",buf[1]);
			assert(write(fd,buf,2)==2);break;
		case 0xE0:
			printf("Channel %i: ",(buf[0]&0x0f)+1);
			printf("Pitch Wheel: 0x%X 0x%X\n",buf[1],buf[2]);
			assert(write(fd,buf,3)==3);break;
		default:
			assert(!"Internal Error, unknown channelcommand.");
	}
}
*/

static void handle_random(unsigned char *buf) {
//nothing done
}

static void midi_put(unsigned char c) {
	static unsigned char buf[3];
	static int next = 0;
	static int needed = 0;
	if(needed) {
		buf[next++] = c;
		if(!(--needed)) {
			if(buf[0]<0xf0)
				handle_channelop(buf);
			else
				handle_random(buf);
			next = 0;
		}
		return;
	}
	if(c<0x80) {
		fprintf(stderr,"Unexpected data: 0x%X\n",c);
		return;
	}
	buf[next++] = c;
	if(c < 0xf0) {
		if(c< 0xC0 || c >= 0xe0) 
			needed = 2;
		else needed = 1;
		return;
	}
	switch (c) {
		case 0xf0: 
			assert(!"Internal error, sysex got through");return;
		case 0xf1:
		case 0xf2:
			needed = 2; return;
		case 0xf3:
			needed = 1;return;
		case 0xf6:
			handle_random(buf); return;
		case 0xf7:
			return;
		default:
			printf("Unknown Status: 0x%X\n",c);
	}
}


static void midi_poll() {
	unsigned char buf[256];
	static bool sysex = false;
	int r = read(midifd,buf,256);
	//if(!r) return;
	for(int i=0;i<r;i++) {
		if(buf[i]>=0xF8) {
			//handle_realtime(buf[0]);
			//discard realtime
			continue;
		}
		if(buf[i]==0xF0) {
			sysex=true;
			continue;
		}
		if(buf[i]<0x80) {
			if(!sysex)
				midi_put(buf[i]);
			else {
				//	fprintf(stderr,"sysex pruning 0x%X\n",buf[i]);
			}
			continue;
		}
		if(buf[i]>=0x80) {
			sysex = false;
			if(buf[i]!=0xf7)
				midi_put(buf[i]);
		}
	}


}


/*
void read_midi(int fd)
{
  unsigned char ev[4];
  int i,n;
  //printf("reading from %i\n",fd);


  if((n=read(seqfd, ev, sizeof(ev)))==-1) {
       printf("Error reading " SEQUENCER_PATH);
       exit(-1);
  }
  for(i=0; i<=(n/4); i++) {
    unsigned char *p = &ev[i*4];

    if(p[0]==SEQ_MIDIPUTC && p[2]==0) {
      if(p[1] & 0x80) { // Status 
	if(bufp) do_buf();
	buf[0] = p[1]; bufp=1;
      } else
	if(bufp) {
	  buf[bufp++] = p[1];
	  if ((buf[0] & 0xf0) == 0x90 || (buf[0] & 0xf0) == 0x80 
	      || (buf[0] & 0xf0) == 0xb0 || (buf[0] & 0xf0) == 0xe0) {
	    if(bufp==3) do_buf();
	  } else if ((buf[0] & 0xf0) == 0xc0 || (buf[0] & 0xf0) == 0xd0) {
	    if(bufp==2) do_buf();	    
	  };
	    
	};
    };
  };
};*/




void init_seq() {
	init_midi();
	init_sequencer();
	init_lists();
	
}


gboolean midi_idle(gpointer data) {
	void reload();
	reload();
	//read_midi(seqfd);
	midi_poll();
	g_idle_add(midi_idle,0);
	return FALSE;
}

